#
# PST module Matrix - put multiple pages on one page
# 
# (c) 2002  Tomas Valla   <valla@atrey.karlin.mff.cuni.cz>
# 
# This software may be used and distributed under the terms
# of the GNU General Public License. See COPYING for details.
#
#
#
#
# options:
# m|modulo = <int>
# w|width = <unit>
# h|height = <unit>
# p|paper = <paper>
# g|grid = (rowlist)
#     where rowlist is [columnlist1] [columnlist2][collist3]....
#           where columnlist is in form page1,page2,....
# 
# it reads modulo pages from input and arranges them in matrix 
# according to grid



package Matrix;
use OM;
use sizes;
@ISA = qw(OM);
use strict;


sub new {
	my $proto = shift;
	my $class = ref($proto) || $proto;
	my $self = OM->new;
	
	$self->name("Matrix");
	$self->{GRID} = {};
	$self->{MODULO} = 1;
	$self->{WIDTH} = undef;
	$self->{HEIGHT} = undef;
	$self->{ROWS} = 0;
	$self->{COLUMNS} = {};
	$self->{ROW_HEIGHT} = undef;
	$self->{COL_WIDTH} = {};
	$self->{PAGE_DESC} = {};

	return bless($self,$class);
}

sub parse_options {
	my $self = shift;
	my $str = "";

	$self->log(4,"parsing options");
	while (my $s = shift) { $str .= $s }

	foreach my $e (split /;/, $str) {
		$e =~ s/\s//g;
		if ($e =~ /^(m|modulo)\=(\d+)$/) {
			$self->{MODULO} = $2;
		} elsif ($e =~ /^(w|width)\=(\w+)$/) {
			my $u = unit_size($2);
			$self->die($u) if $u =~ /^ERROR/;
			$self->{WIDTH} = $u;
		} elsif ($e =~ /^(h|height)\=(\w+)$/) {
			my $u = unit_size($2);
			$self->die($u) if $u =~ /^ERROR/;
			$self->{HEIGHT} = $u;
		} elsif ($e =~ /^(p|paper)\=(\w+)$/) {
			my @r = paper_size($2);
			$self->die($r[0]) if $r[0] =~ /^ERROR/;
			($self->{WIDTH},$self->{HEIGHT}) = @r;
		} elsif ($e =~ /^(g|grid)\=\(([^\)]*)\)$/) {
			my $row = 0;
			my $q = $2;

			$q =~ s/\]\[/\]%\[/g;
			foreach my $ln (split /%/, $q) {
				if ($ln !~ /^\[((\d+,)*(\d+)?)\]$/) {
					$self->die("invalid linespec ($ln)\n");
				}
				++$row;
				my $col = 0;
				foreach my $pg (split /,/, $1) {
					++$col;
					$self->{GRID}{$row}{$col} = $pg;
				}
				$self->{COLUMNS}{$row} = $col;
			}
			$self->{ROWS} = $row;
		} else {
			$self->die("bad option ($e)\n");
		}
	}
	if(!defined $self->{WIDTH} || !defined $self->{HEIGHT}) {
		$self->die("I need some option\n");
	}
	for (my $r = 1; $r <= $self->{ROWS}; $r++) {
		for (my $c = 1; $c <= $self->{COLUMNS}{$r}; $c++) {
			my $x = $self->{GRID}{$r}{$c};
			$self->die("invalid page index in grid") if $x<0 || $x>=$self->{MODULO};
		}
	}
	$self->{ROW_HEIGHT} = $self->{HEIGHT} / $self->{ROWS};
	foreach my $k (keys %{$self->{COLUMNS}}) {
		$self->{COL_WIDTH}{$k} = $self->{WIDTH} / $self->{COLUMNS}{$k};
	}
}

sub page_size {
	my $self = shift;
	my $page = shift;
	my $pp = ($page-1)*$self->{MODULO};
	my $flag = 0;
	my $f = $self->{FILE_HANDLER};
	my $pgdesc = "(";

	foreach my $ln (values %{$self->{GRID}}) {
		foreach my $pg (values %{$ln}) {
			my $cpg = $pp+$pg+1;
			if ($cpg <= $self->{PREV_PAGES}) {
				$flag = 1;
				$pgdesc .= ($pgdesc eq "(" ? "" : ",") . $f->{PAGE_NUMBERS}{$cpg};
			}
		}
	}
	return 0 if not $flag;
	$pgdesc .= ")";
	$self->{PAGE_DESC}{$page} = $pgdesc;
	$self->{PAGE_WIDTH}{$page} = $self->{WIDTH};
	$self->{PAGE_HEIGHT}{$page} = $self->{HEIGHT};
	return 1;
}

sub run {
	my $self = shift;

	$self->{PREVIOUS}->run;
	for (my $i = 1; $i <= $self->{NUM_PAGES}; $i++) {
		$self->{FILE_HANDLER}->{PAGE_NUMBERS}{$i} = $self->{PAGE_DESC}{$i};
	}
}


sub write_page_prolog {
	my $self = shift;
	my $f = $self->{OUTPUT};

	print $f
		"userdict/PSTdict save put\n";
}

sub write_page_epilog {
	my $self = shift;
	my $f = $self->{OUTPUT};

	print $f
		"PSTdict restore showpage\n";
}

sub put_page {
	my $self = shift;
	my $page = shift;
	my ($llx,$lly,$urx,$ury) = @_;
	my $f = $self->{OUTPUT};

	my $xs = ($urx-$llx) / $self->{PREV_PAGE_WIDTH}{$page};
	my $ys = ($ury-$lly) / $self->{PREV_PAGE_HEIGHT}{$page};
	my $x = $llx;
	my $y = $lly;

	print $f "userdict/PSTdict save put\nPSTmatrix setmatrix\n",
		"$x $y translate $xs $ys scale\n",
	        "userdict/PSTmatrix matrix currentmatrix put\n",
                #"userdict/PSTclip{0 0 moveto\n",
                #$pw," 0 rlineto 0 ",$ph," rlineto -",$pw," 0 rlineto\n",
                #"closepath}put initclip\n",
		"/showpage{}def/copypage{}def/erasepage{}def\n",
                "PSTxform concat\n";


	$self->previous->write_page($page);
	print $f "PSTdict restore\n";
}



sub write_page {
	my $self = shift;
	my $page = shift;
	my $mdl = $self->{MODULO};

	$self->write_page_prolog($page);
	foreach my $r (1..$self->{ROWS}) {
		foreach my $c (1..$self->{COLUMNS}{$r}) {
			my $rh = $self->{ROW_HEIGHT};
			my $cw = $self->{COL_WIDTH}{$r};
			my $pp = ($page - 1)*$mdl + $self->{GRID}->{$r}{$c} +1;
			if ($pp <= $self->{PREV_PAGES}) {
				$self->put_page($pp,
					($c-1)*$cw,($self->{ROWS}-$r)*$rh,
					$c*$cw,($self->{ROWS}-$r)*$rh+$rh);
			}
		}
	}
	$self->write_page_epilog($page);
}


1; # no init
