#
# PST lib - Basic class for operation modules
# 
# (c) 2002  Tomas Valla   <valla@atrey.karlin.mff.cuni.cz>
# 
# This software may be used and distributed under the terms
# of the GNU General Public License. See COPYING for details.
#

package OM;
use strict;
use IO::File;


sub new {
	my $proto = shift;
	my $class = ref($proto) || $proto;
	my $self = {};
	$self->{VERBOSE} = 0;
	$self->{PREV_PAGE_WIDTH} = {};    # this contains page size as reported by previous module
	$self->{PREV_PAGE_HEIGHT} = {};
	$self->{PAGE_WIDTH} = {};	# page sizes produced by this module
	$self->{PAGE_HEIGHT} = {};
	$self->{PAGE_H_OFF} = {};
	$self->{PAGE_V_OFF} = {};   # page offsets
	$self->{PREV_PAGES} = undef;  # number of previous pages
	$self->{NUM_PAGES} = undef;   # my page count
	$self->{OUTPUT} = undef;      # filedesriptor for easy output writing
	$self->{FILE_HANDLER} = undef;  # input file DSC object
	$self->{PREVIOUS} = undef;      # previous module
	$self->{NAME} = "OM";

	return bless($self,$class);
}


sub DESTROY {
	my $self = shift;
	$self->output->close if defined $self->output;
}


# simple logging stuff
sub log {
	my $self = shift;
	my $level = shift;

	print $self->name,": ",@_,"\n"   if $level <= $self->{VERBOSE};
}

sub die {
	my $self = shift;

	print $self->name," cannot run: ",@_,"\n";
	print "Terminating...\n";
	exit(1);
}




# some encapsulation...

sub verbose {
	my $self = shift;
	if(@_) { $self->{VERBOSE} = shift }
	return $self->{VERBOSE};
}

sub name {
	my $self = shift;
	if(@_) { $self->{NAME} = shift }
	return $self->{NAME};
}

sub output {
	my $self = shift;
	if(@_) { $self->{OUTPUT} = shift }
	return $self->{OUTPUT};
}

sub previous {
	my $self = shift;
	if(@_) { $self->{PREVIOUS} = shift }
	return $self->{PREVIOUS};
}



# now some methods intended to be redefined in derived classes

# init should contain code which is executed in the beginning
# only parse_options is called earlier
sub init {
	my $self = shift;
	defined $self->previous && $self->previous->init;
	$self->{OUTPUT} = $self->previous->output;
	$self->{FILE_HANDLER} = $self->previous->{FILE_HANDLER};
}

# try to compute page sizes for all pages produced by module
# it tries to produce pages starting with 1 by calling page_size
# if page_size returns 0, it is signal no more pages will be produced
sub calculate {
	my $self = shift;

	return  if not defined $self->previous;
	my $p = $self->previous;
	$p->calculate;
	for(my $i=1; $i<=$p->{NUM_PAGES}; $i++) {
		$self->{PREV_PAGE_WIDTH}{$i} = $p->{PAGE_WIDTH}{$i};
		$self->{PREV_PAGE_HEIGHT}{$i} = $p->{PAGE_HEIGHT}{$i};
	}
	$self->{PREV_PAGES} = $p->{NUM_PAGES};

	my $i = 1;
	while ($self->page_size($i)) {   # expect true if page $i will be produced, otherwise end
		++ $i;
	}
	$self->{NUM_PAGES} = $i - 1;
}

# code executed after calculate
sub run {
	my $self = shift;

	defined $self->previous && $self->previous->run;
}

# done should contain code to be executed at the end of all
sub done {
	my $self = shift;
	defined $self->previous && $self->previous->done;
}







# this method should be redefined in derived classes
# default behaviour is to copy page size from preceeding module
sub page_size {
	my $self = shift;
	my $page = shift;

	$self->log(5, "page_size start\n");
	if ($page > 0  &&  $page <= $self->{PREV_PAGES}) {
		$self->{PAGE_WIDTH}{$page} = $self->{PREV_PAGE_WIDTH}{$page};
		$self->{PAGE_HEIGHT}{$page} = $self->{PREV_PAGE_HEIGHT}{$page};
		$self->log(5,"page_size end (success)\n");
		return 1;
	}
	$self->log(5,"page_size end (fail)\n");
	return 0;
}


# default behaviour - write prolog, call previous module, write epilog
sub write_page {
	my $self = shift;
	my $page = shift;

	$self->write_page_prolog($page);
	$self->previous->write_page($page);
	$self->write_page_epilog($page);
}



# these three should be redefined in derived classes
sub write_page_prolog {
}
sub write_page_epilog {
}
sub parse_options {
}

1;  # no initialization...

