/*
 * psscan - calculate positions of Postscript DSC comments
 * 
 * (c) 2002  Tomas Valla   <valla@atrey.karlin.mff.cuni.cz>
 * 
 * This software may be used and distributed under the terms
 * of the GNU General Public License. See COPYING for details.
 *
 *
 *
 *  read stdin and writes informations to stdout in form
 *     byteno  comment
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>



/*
 * We use our own buffered input because it is much faster
 * than stdio and it allows tell() even on pipes
 */

static unsigned char inbuf[16384];
static unsigned char *inptr = inbuf;
static unsigned char *inend = inbuf;
static int inpos = 0;

/* getchar, position info, ungetchar */
#define gc() ((inptr < inend) ? *inptr++ : slow_gc())
#define tell() (inpos + (inptr - inbuf))
#define ungc() inptr--

static int slow_gc(void)
{
	int len = read(0, inbuf, sizeof(inbuf));
	inpos += inend - inbuf;
	if (len < 0) {
		fprintf(stderr, "read: %s\n", strerror(errno));
		exit(1);
	}
	inptr = inbuf;
	inend = inbuf + len;
	if (!len)
		return EOF;
	return *inptr++;
}

int main(int argc, char **argv)
{
	int c, skip, skip_lines;
	unsigned pos;
	char buf[256], *b;
	int inside = 0;
	int comments = 1;
	int mf;

	/* Find "%!PS" comment */
	for (;;) {
		c = gc();
		if (c == EOF) {
			fprintf(stderr, "Not a structured PostScript file.\n");
			return 1;
		}
		if (c != '%' && (c <= ' ' || c >= 0x7f))
			continue;
		if (c == '%' && (c = gc()) == '!' && (c = gc()) == 'P' && (c = gc()) == 'S')
			break;
		while (c >= ' ' && c <= 0x7f)
			c = gc();
	}
	printf("%d %%!PS", tell() - 4);
	while ((c = gc()) != EOF && c != '\r' && c != '\n')
		putchar(c);
	putchar('\n');

	/* Find all DSC comments */
	while (c != EOF) {
		pos = tell() - 1;
		mf = 0;
		if (c != '%') {
			mf = (c != '\n' && c != '\r');
		      skiptoeol:
			if (inside) {
				putchar('\n');
				inside = 0;
			}
			if (comments && mf) {
				/* Supply %%EndComments if header ends prematurely */
				comments = 0;
				printf("%d %%%%EndComments\n", pos);
			}
			while (c != '\n' && c != '\r' && c != EOF)
				c = gc();
			c = gc();
			continue;
		}
		c = gc();
		if (c != '%')
			goto skiptoeol;
		c = gc();
		if (c == '+') {
			if (!inside)
				goto skiptoeol;
			c = gc();
			if (c == '\r' || c == '\n' || c == EOF)
				goto skiptoeol;
		} else if (!((c >= 'A' && c <= 'Z') || (c >= 'a' && c <= 'z')))
			goto skiptoeol;
		else {
			if (inside)
				putchar('\n');
			printf("%d %%%%", pos);
		}
		inside = 1;
		b = buf;
		do {
			if (b < buf + sizeof(buf) - 1)
				*b++ = c;
			putchar(c);
			c = gc();
		}
		while (c != '\n' && c != '\r' && c != EOF);
		*b = 0;
		skip = skip_lines = 0;

		/*
		 * We need to find the binary parts of the file and skip them.
		 * This is the reason of input line buffering.
		 */
		if (!strncmp(buf, "BeginBinary", 11)) {
			b = buf + 11;
			skip = strtoul(b, NULL, 10);
		}
		if (!strncmp(buf, "BeginData", 9)) {
			b = buf + 9;
			skip = strtoul(b, NULL, 10);
			if (strstr(b, "Lines"))
				skip_lines = 1;
		}
		if (!strncmp(buf, "EndComments", 11))
			comments = 0;
		if (skip) {
			/*
			 * DSC specification is very unclear about binary data. It allows
			 * various line endings (CR,LF,CRLF) but doesn't say how to treat
			 * them in binary data. I assume that BeginBinary/BeginData are
			 * separate lines and I skip their EOL before data counting.
			 * I assume that all lines in binary data have the same EOL.
			 */
			if (c == '\r') {
				c = gc();
				if (c != '\n')
					ungc();
			}
			if (skip_lines) {
				while (skip--) {
					do
						c = gc();
					while (c != EOF && c != '\r' && c != '\n');
					if (c == '\r') {
						c = gc();
						if (c != '\n')
							ungc();
					}
				}
			} else
				while (c != EOF && skip--)
					c = gc();
		}
		c = gc();
	}
	if (inside)
		putchar('\n');
	printf("%d %%%%EOF\n", tell());  /* Just to be sure */

	return 0;
}
